@@ -1,7 +1,7 @@
 #!/usr/bin/perl -T
 ###
-###              analog 4.16             http://www.analog.cx/
-### This program is copyright (c) Stephen R. E. Turner 1995 - 2001 except as
+###              analog 5.32             http://www.analog.cx/
+### This program is copyright (c) Stephen R. E. Turner 1995 - 2003 except as
 ### stated otherwise. Distribution, usage and modification of this program is
 ### subject to the conditions of the Licence which you should have received
 ### with it. This program comes with no warranty, expressed or implied.
@@ -13,8 +13,8 @@
 
 # 1) uncomment (remove everything before $analog) and edit one of the next two
 #    lines to give the location (full pathname) of the analog executable.
-#       Unix: $analog = '/usr/local/etc/httpd/analog-4.16/analog';
-#       Windows: $analog = 'C:\program files\analog 4.16\analog.exe';
+#       Unix: $analog = '/usr/local/etc/httpd/analog-5.32/analog';
+#       Windows: $analog = 'C:\program files\analog 5.32\analog.exe';
 # 2) If you're on Unix, edit the first line in this file to give the location
 #    of Perl (don't remove the #! though).
 # 3) You also need to edit anlgform.html if you want to use the form.
@@ -22,18 +22,25 @@
 
 @forbidden = qw(LOGFORMAT APACHELOGFORMAT DEFAULTLOGFORMAT
 		APACHEDEFAULTLOGFORMAT HEADERFILE FOOTERFILE UNCOMPRESS
-		OUTFILE CACHEOUTFILE ERRFILE DNS CGI SETTINGS LANGFILE);
+		OUTFILE CACHEOUTFILE LOCALCHARTDIR ERRFILE DNS CGI
+		SETTINGS PROGRESSFREQ LANGFILE DESCFILE);
 # Forbidden commands: sysadmin can add more (must be in upper case!)
 # Other commands you might consider adding, because they allow users to
 # specify which files to use for the analysis, are LOGFILE and DOMAINSFILE.
 # If you add a command, you must also add any aliases it possesses.
 # There is a discussion of all this in docs/form.html.
 
+@allowed = qw();
+# Allowed commands. If there are _any_ commands listed here, then _only_
+# commands which are in @allowed, and not in @forbidden, can be used.
+
 require 5.001;
 use CGI;
 
 # 1) INITIALISATION
-$ENV{PATH} = '';
+# delete all dangerous environment variables
+$ENV{PATH} = '';  # blank, not deleted, so that UNCOMPRESS doesn't get a path
+delete @ENV{qw/IFS CDPATH ENV BASH_ENV/};
 $query = new CGI;
 $|=1;
 $lt = localtime;
@@ -60,7 +67,7 @@ checkfilechars("CACHEFILE");
 
 # 2) OPEN THE ANALOG PROCESS
 # qv=1 causes args to go straight to stdout, not program
-if (${$args{'QV'}}[-1] && !grep('QV' eq $_, @forbidden)) {
+if (${$args{'QV'}}[-1] && !forbidden('QV')) {
     print "Content-Type: text/plain\n\n";
     open(ANALOG, ">-");
 }
@@ -85,21 +92,23 @@ else {
 
 # 3) PRINT ALL THE COMMANDS
 # Special cases: must come first
-printargs('CG', 'CONFIGFILE') unless grep('CG' eq $_, @forbidden);
+printargs('CG', 'CONFIGFILE') unless forbidden('CG');
+# both 'CG' and 'CONFIGFILE' must be allowed for this to happen.
 print ANALOG "CGI ON\nDNS NONE\nWARNINGS FL\n";
-printargs('WARNINGS') unless grep('WARNINGS' eq $_, @forbidden);
-printargs('LOGTIMEOFFSET') unless grep('LOGTIMEOFFSET' eq $_, @forbidden);
+printargs('WARNINGS');
+printargs('LOGTIMEOFFSET');
 foreach $k (@order) {
     printargs($k)
         unless($k eq 'QV' || $k eq 'CG' || $k eq 'CM' || $k =~ /FLOORB$/ ||
 	       $k =~ /2$/ || $k =~ '^LOGTIMEOFFSET' || $k =~ '^WARNINGS' ||
                # commands dealt with elsewhere
-               $k =~ /[^A-Z12]/ || $k eq '' ||
-	       grep($k =~ /^$_.?$/, @forbidden) ||
-	       $k =~ /^IGNORE/); # other stuff not wanted
+               $k =~ /[^A-Z12]/ || $k eq '' || $k =~ /^IGNORE/);
+               # other stuff not wanted
 }
-# Special case: must come last
-printargs('CM', 'CONFIGFILE') unless grep('CM' eq $_, @forbidden);
+# Special cases: must come last
+print ANALOG "DEBUG -C\n";
+printargs('CM', 'CONFIGFILE') unless forbidden('CM');
+# again, both 'CM' and 'CONFIGFILE' must be allowed for this to happen.
 print ANALOG "OUTFILE stdout\n";
 
 # 4) WAIT FOR PROCESS TO FINISH. THAT'S IT.
@@ -113,7 +122,15 @@ unless (close(ANALOG)) {
     die;
 }
 
-# A) PRINT ONE COMMAND
+
+### SUBROUTINES 
+# A) IS A GIVEN COMMAND FORBIDDEN?
+sub forbidden {
+    return (grep($_[0] eq $_, @forbidden) ||
+	    (@allowed && !grep($_[0] eq $_, @allowed)));
+}
+
+# B) PRINT ONE COMMAND
 sub printargs {
     my($is_floora) = 0;
     my($is_12) = 0;
@@ -121,23 +138,29 @@ sub printargs {
     if ($name =~ /FLOORA$/) {
 	chop($name);
 	$is_floora = 1;
-	$b = ${$args{$name . 'B'}}[-1];
     }
     elsif ($name =~ /1$/) {
 	chop($name);
 	$is_12 = 1;
-	$b = ${$args{$name . '2'}}[-1];
     }
-    foreach $a (@{$args{$_[0]}}) {
-	if ($a ne '') {
-	    if ($is_floora) {
-		print ANALOG ("$name $a$b\n") if ($b ne '');
-	    }
-	    elsif ($is_12) {
-		print ANALOG ("$name ", bracket($a), " ", bracket($b), "\n")
-		    if ($b ne '');
-	    }
-	    else {
+    return if forbidden($name);
+
+    if ($is_floora) {
+	$a = ${$args{$name . 'A'}}[-1];  # last "FLOORA=$a" form arg specified
+	$b = ${$args{$name . 'B'}}[-1];
+	print ANALOG ("$name $a$b\n") if ($b ne '' && $b !~ /\\$/);
+	# could bracket $a$b, but no help because any special character in a
+	# FLOOR command is junk anyway.
+    }
+    elsif ($is_12) {
+	$a = ${$args{$name . '1'}}[-1];
+	$b = ${$args{$name . '2'}}[-1];
+	print ANALOG ("$name ", bracket($a), " ", bracket($b), "\n")
+	    if ($b ne '');
+    }
+    else {
+	foreach $a (@{$args{$_[0]}}) {  # run through all "NAME=$a" form args
+	    if ($a ne '') {
 		print ANALOG ("$name ", bracket($a), "\n");
 		print ANALOG ("DNS READ\n") if ($name eq 'DNSFILE');
 	    }
@@ -145,21 +168,23 @@ sub printargs {
     }
 }
 
-# B) PUT APPROPRIATE DELIMITERS ROUND AN ARGUMENT CONTAINING SPACES
+# C) PUT APPROPRIATE DELIMITERS ROUND AN ARGUMENT CONTAINING SPACES
 sub bracket {
     local $_ = $_[0];
-    return $_ unless (/[\s\#]/ || /^['"\(]/);
+    return $_ unless (/[\s\#]/ || /^['"\(]/ || /\\$/);
     return "\"$_\"" unless (/"/);
     return "'$_'" unless (/'/);
     return "($_)";
     # analog has no syntax if string contains ) as well as space, ' and "
 }
 
-# C) CHECK ONLY SAFE CHARACTERS in LOGFILEs and CACHEFILEs. See docs/form.html.
+# D) CHECK ONLY SAFE CHARACTERS in LOGFILEs and CACHEFILEs. See docs/form.html.
 sub checkfilechars {
     local ($_);
     foreach (@{$args{$_[0]}}, @{$args{$_[0] . '1'}}) {
-	if (m([^\w\. /\\:-]) || m(\W-|^-|-\W|-$)) { # NB \w includes underscore
+	if (m([^\w\. /\\:\-\*\?]) || m(\B-|-\B)) {
+	    # i.e. contains a non-approved character, or a dash not
+	    # between \w's. NB \w includes underscore.
 	    badreq(403, "Illegal Request", "Unsafe characters in $_[0].");
 	    printf STDERR "[$lt] $progname: Unsafe characters in \"$_[0] $_\" on request from %s\n", $ENV{REMOTE_HOST}?$ENV{REMOTE_HOST}:($ENV{REMOTE_ADDR}?$ENV{REMOTE_ADDR}:"unknown host");
 	    die;
@@ -167,17 +192,22 @@ sub checkfilechars {
     }
 }
 
-# D) CHECK ONLY SAFE CHARACTERS IN OTHER COMMANDS. Again, see docs/form.html.
+# E) CHECK NO UNSAFE CHARACTERS IN OTHER COMMANDS. Again, see docs/form.html.
 sub checkchars {
     local $_ = $_[0];
     if (/[\x00-\x1F\x7F-\x9F]/) {
-	badreq(403, "Illegal Request", "Unsafe characters in \U$p.");
 	printf STDERR "[$lt] $progname: Unsafe characters in \"\U$p\E $_\" on request from %s\n", $ENV{REMOTE_HOST}?$ENV{REMOTE_HOST}:($ENV{REMOTE_ADDR}?$ENV{REMOTE_ADDR}:"unknown host");
+	# Translate dangerous characters to avoid cross-site scripting
+	$p =~ s/&/&amp;/;
+	$p =~ s/</&lt;/;
+	$p =~ s/>/&gt;/;
+	$p =~ s/"/&quot;/;
+	badreq(403, "Illegal Request", "Unsafe characters in \U$p.");
 	die;
     }
 }
 
-# E) PRINT OUT ERROR MESSAGE
+# F) PRINT OUT ERROR MESSAGE
 sub badreq {
     my($i);
     print "Content-Type: text/html\n";
